/**
 *
 * Magnolia Criteria API (http://www.openmindlab.com/lab/products/mgnlcriteria.html)
 * Copyright(C) 2009-2010, Openmind S.r.l. http://www.openmindonline.it
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.sourceforge.openutils.mgnlcriteria.advanced;

import info.magnolia.cms.beans.config.ContentRepository;
import info.magnolia.cms.core.Content;
import info.magnolia.context.MgnlContext;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import net.sourceforge.openutils.mgnlcriteria.jcr.query.AdvancedResult;
import net.sourceforge.openutils.mgnlcriteria.jcr.query.AdvancedResultItem;
import net.sourceforge.openutils.mgnlcriteria.jcr.query.Criteria;
import net.sourceforge.openutils.mgnlcriteria.jcr.query.JCRCriteriaFactory;
import net.sourceforge.openutils.mgnlcriteria.jcr.query.JCRQueryException;
import net.sourceforge.openutils.mgnlcriteria.jcr.query.ResultIterator;
import net.sourceforge.openutils.mgnlcriteria.jcr.query.criterion.Disjunction;
import net.sourceforge.openutils.mgnlcriteria.jcr.query.criterion.Order;
import net.sourceforge.openutils.mgnlcriteria.jcr.query.criterion.Restrictions;
import net.sourceforge.openutils.mgnlcriteria.tests.RepositoryTestNgTestcase;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;
import org.testng.Assert;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;


/**
 * @author lbrindisi
 * @version $Id: AdvancedCriteriaSearchTest.java 2838 2010-08-03 11:01:57Z fgiust $
 */
public class AdvancedCriteriaSearchTest extends RepositoryTestNgTestcase
{

    /**
     * {@inheritDoc}
     */
    @Override
    @BeforeClass
    protected void setUp() throws Exception
    {
        setRepositoryConfigFileName("/crit-repository/test-repositories.xml");
        setJackrabbitRepositoryConfigFileName("/crit-repository/jackrabbit-test-configuration.xml");

        super.setUp();

        bootstrapSingleResource("/crit-bootstrap/website.00000.xml");
        MgnlContext.getHierarchyManager(ContentRepository.WEBSITE).save();
    }

    @Test
    public void testSearchFrancia() throws Exception
    {
        // Frància
        // Franciacorta
        // Parigi (Francia)
        // Parigi (Frància)
        // federale -> francia in paragraphs

        AdvancedResult advResult = search("francia", 1, 200);
        Collection< ? extends Content> result = collectCollectionFromResult(advResult);

        // "frappè" should not show up since "excludedproperty" is not indexed. But looks like it only works with
        // jackrabbit 2?

        assertSortedResults(
            new String[]{"Frància", "Parigi (Francia)", "Parigi (Frància)", },
            ((List< ? extends Content>) result).subList(0, 3),
            "francia");

        // the remaining 4 pages have the same score, so the order is not stable (changes between jdk5 and jdk6)
        // "Tallart, Camille d'Hostun, cónte di-",
        // "federale",
        // ":)",
        // "Faccina sorridente :)"
    }

    @Test
    public void testSearchFranciaPaginated() throws Exception
    {

        AdvancedResult advResult = search("francia", 3, 1);

        Assert.assertEquals(advResult.getTotalSize(), 7);
        Collection< ? extends Content> result = collectCollectionFromResult(advResult);

        assertSortedResults(new String[]{"Parigi (Frància)" }, result, "francia");
    }

    @Test
    public void testSearchDante() throws Exception
    {
        // Dante Alighieri
        // Alighieri, Dante

        AdvancedResult advResult = search("Dante Alighieri", 1, 200);
        Collection< ? extends Content> result = collectCollectionFromResult(advResult);

        assertNumOfResults(2, result, "Dante Alighieri");
    }

    @Test
    public void testSearchFagiano() throws Exception
    {
        // fagiano
        // fagiàno
        // fàgiànò

        AdvancedResult advResult = search("fagiano", 1, 200);
        Collection< ? extends Content> result = collectCollectionFromResult(advResult);

        assertNumOfResults(3, result, "fagiano");
    }

    @Test
    public void testSearchPesca() throws Exception
    {
        // pèsca
        // canna da pesca

        AdvancedResult advResult = search("pesca", 1, 200);
        Collection< ? extends Content> result = collectCollectionFromResult(advResult);

        assertSortedResults(new String[]{"pèsca", "canna da pesca" }, result, "pesca");
    }

    @Test
    public void testEscapeQuotesForEqRestriction() throws Exception
    {
        String title = "Tallart, Camille d'Hostun, cónte di-";
        Criteria criteria = JCRCriteriaFactory.createCriteria().setWorkspace(ContentRepository.WEBSITE);
        criteria.setBasePath(StringUtils.EMPTY);
        criteria.add(Restrictions.eq("@jcr:primaryType", "mgnl:content"));
        criteria.add(Restrictions.eq("@title", title));
        AdvancedResult advResult = null;
        try
        {
            advResult = criteria.execute();
        }
        catch (JCRQueryException e)
        {
            Assert.fail("Invalid query");
        }
        Assert.assertNotNull(advResult);
        Collection< ? extends Content> collection = collectCollectionFromResult(advResult);
        Assert.assertEquals(collection.size(), 1);
        for (Content content : collection)
        {
            Assert.assertEquals(content.getTitle(), title);
        }
    }

    @Test
    public void testEscapeDoubleQuotesForContainsRestriction() throws Exception
    {
        String searchText = "\"Milano\"";
        Criteria criteria = JCRCriteriaFactory.createCriteria().setWorkspace(ContentRepository.WEBSITE);
        criteria.setBasePath(StringUtils.EMPTY);
        criteria.add(Restrictions.eq("@jcr:primaryType", "mgnl:content"));
        criteria.add(Restrictions.contains("@title", searchText));
        try
        {
            criteria.execute();
        }
        catch (JCRQueryException e)
        {
            Assert.fail("Search string not properly escaped. " + e.getMessage());
        }
    }

    @Test
    public void testEscapeInvalidChars() throws Exception
    {
        String searchText = "\"Milano(){}[]+*?^|\\/!";
        Criteria criteria = JCRCriteriaFactory.createCriteria().setWorkspace(ContentRepository.WEBSITE);
        criteria.setBasePath(StringUtils.EMPTY);
        criteria.add(Restrictions.eq("@jcr:primaryType", "mgnl:content"));
        criteria.add(Restrictions.contains("@title", searchText));
        try
        {
            criteria.execute();
        }
        catch (JCRQueryException e)
        {
            Assert.fail("Search string not properly escaped. " + e.getMessage());
        }
    }

    @Test
    public void testEscapeSingleQuotesForContainsRestriction() throws Exception
    {
        String searchText = "Milano'";
        Criteria criteria = JCRCriteriaFactory.createCriteria().setWorkspace(ContentRepository.WEBSITE);
        criteria.setBasePath(StringUtils.EMPTY);
        criteria.add(Restrictions.eq("@jcr:primaryType", "mgnl:content"));
        criteria.add(Restrictions.contains("@title", searchText));
        try
        {
            criteria.execute();
        }
        catch (JCRQueryException e)
        {
            Assert.fail("Search string not properly escaped. " + e.getMessage());
        }
    }

    @Test
    public void testEscapePlusForContainsRestriction() throws Exception
    {
        String searchText = "Milano +";
        Criteria criteria = JCRCriteriaFactory.createCriteria().setWorkspace(ContentRepository.WEBSITE);
        criteria.setBasePath(StringUtils.EMPTY);
        criteria.add(Restrictions.eq("@jcr:primaryType", "mgnl:content"));
        criteria.add(Restrictions.contains("@title", searchText));
        try
        {
            criteria.execute();
        }
        catch (JCRQueryException e)
        {
            Assert.fail("Search string not properly escaped. " + e.getMessage());
        }
    }

    @Test
    public void testNotInPrivatePropertiesNoMagnoliaUser()
    {

        AdvancedResult advResult = search("superuser", 1, 200);
        Collection< ? extends Content> result = collectCollectionFromResult(advResult);

        assertNumOfResults(0, result, "superuser");
    }

    @Test
    public void testNotInPrivatePropertiesNoTemplates()
    {
        AdvancedResult advResult = search("t-redirect", 1, 200);
        Collection< ? extends Content> result = collectCollectionFromResult(advResult);

        assertNumOfResults(0, result, "t-redirect");
    }

    @Test
    public void testEscapeReservedKeyword() throws Exception
    {
        String searchText = "OR SONO";
        Criteria criteria = JCRCriteriaFactory.createCriteria().setWorkspace(ContentRepository.WEBSITE);
        criteria.setBasePath(StringUtils.EMPTY);
        criteria.add(Restrictions.eq("@jcr:primaryType", "mgnl:content"));
        criteria.add(Restrictions.contains("@title", searchText));
        try
        {
            AdvancedResult advResult = criteria.execute();
            assertNumOfResults(1, collectCollectionFromResult(advResult), searchText);
        }
        catch (JCRQueryException e)
        {
            Assert.fail("Search string not properly escaped. " + e.getMessage());
        }
    }

    @Test
    public void testEscapePipes() throws Exception
    {
        String searchText = "giovanni paolo ||";
        Criteria criteria = JCRCriteriaFactory.createCriteria().setWorkspace(ContentRepository.WEBSITE);
        criteria.setBasePath(StringUtils.EMPTY);
        criteria.add(Restrictions.eq("@jcr:primaryType", "mgnl:content"));
        criteria.add(Restrictions.contains("@title", searchText));
        try
        {
            AdvancedResult advResult = criteria.execute();
            assertNumOfResults(1, collectCollectionFromResult(advResult), searchText);
        }
        catch (JCRQueryException e)
        {
            Assert.fail("Search string not properly escaped. " + e.getMessage());
        }
    }

    @Test
    public void testEscapeUnconventionalKeywords() throws Exception
    {
        String searchText = "(ai:)(n)(uk)";

        Criteria criteria = JCRCriteriaFactory.createCriteria().setWorkspace(ContentRepository.WEBSITE);
        criteria.setBasePath(StringUtils.EMPTY);
        criteria.add(Restrictions.eq("@jcr:primaryType", "mgnl:content"));
        criteria.add(Restrictions.contains("@title", searchText));
        try
        {
            AdvancedResult advResult = criteria.execute();
            assertNumOfResults(0, collectCollectionFromResult(advResult), searchText);
        }
        catch (JCRQueryException e)
        {
            Assert.fail("Search string not properly escaped. " + e.getMessage());
        }
    }

    @Test
    public void testEscapeUnconventionalKeywords0() throws Exception
    {
        String searchText = "(fr:)(n)";
        Criteria criteria = JCRCriteriaFactory.createCriteria().setWorkspace(ContentRepository.WEBSITE);
        criteria.setBasePath(StringUtils.EMPTY);
// criteria.add(Restrictions.eq("@jcr:primaryType", "mgnl:content"));
// criteria.add(Restrictions.contains("@title", searchText));

        Disjunction disjunction = Restrictions.disjunction();
        disjunction.add(Restrictions.contains("@title", StringUtils.defaultString(searchText))); // serve per il
        // boost!
        disjunction.add(Restrictions.contains(".", StringUtils.defaultString(searchText)));
        criteria.add(disjunction);

        try
        {
            AdvancedResult advResult = criteria.execute();
            assertNumOfResults(0, collectCollectionFromResult(advResult), searchText);
        }
        catch (JCRQueryException e)
        {
            Assert.fail("Search string not properly escaped. " + e.getMessage());
        }
    }

    @Test
    public void testExceptSmile()
    {
        String searchText = ":)";

        AdvancedResult advResult = search(searchText, 1, 200);
        Collection< ? extends Content> result = collectCollectionFromResult(advResult);

        assertNumOfResults(0, result, searchText); // Probabli it's not indexed or search doesn't work on fulltext

    }

    @Test
    public void testEqualsSmile()
    {
        String searchText = ":)";

        Criteria criteria = JCRCriteriaFactory.createCriteria().setWorkspace(ContentRepository.WEBSITE);
        criteria.setBasePath(StringUtils.EMPTY);
        criteria.add(Restrictions.eq("@jcr:primaryType", "mgnl:content"));
        criteria.add(Restrictions.eq("@title", searchText));
        try
        {
            AdvancedResult advResult = criteria.execute();
            assertNumOfResults(1, collectCollectionFromResult(advResult), searchText);
        }
        catch (JCRQueryException e)
        {
            Assert.fail("Search string not properly escaped. " + e.getMessage());
        }

    }

    protected void assertNumOfResults(int expected, Collection< ? extends Content> result, String search)
    {
        if (result.size() != expected)
        {
            List<String> titles = new ArrayList<String>();
            for (Content content : result)
            {
                titles.add(content.getTitle());
            }

            Assert.fail("Wrong result when searching for \""
                + search
                + "\", expected "
                + expected
                + " results, found "
                + result.size()
                + ". Pages found: "
                + ArrayUtils.toString(titles));
        }
    }

    protected void assertSortedResults(String[] expected, Collection< ? extends Content> result, String search)
    {
        List<String> titles = new ArrayList<String>();
        for (Content content : result)
        {
            titles.add(content.getTitle());
        }

        if (result.size() != expected.length)
        {
            Assert.fail("Wrong result when searching for \""
                + search
                + "\", expected "
                + expected.length
                + " results, found "
                + result.size()
                + ". Pages found: "
                + ArrayUtils.toString(titles));
        }

        Assert.assertEquals(
            ArrayUtils.toString(expected),
            ArrayUtils.toString(titles.toArray()),
            "Wrong order searching for \"" + search + "\"");

    }

    public AdvancedResult search(String searchText, int page, int itemsPerPage)
    {
        return search(searchText, StringUtils.EMPTY, ContentRepository.WEBSITE, false, page, itemsPerPage);
    }

    protected AdvancedResult search(String searchText, String path, String repository, boolean titleOnly, int page,
        int itemsPerPage)
    {
        if (StringUtils.isBlank(searchText))
        {
            return AdvancedResult.EMPTY_RESULT;
        }

        Criteria criteria = createCriteria(repository, path);

        if (titleOnly)
        {
            criteria.add(Restrictions.contains("@title", StringUtils.defaultString(searchText)));
        }
        else
        {
            Disjunction disjunction = Restrictions.disjunction();
            disjunction.add(Restrictions.contains("@title", StringUtils.defaultString(searchText))); // serve per il
            // boost!
            disjunction.add(Restrictions.contains(".", StringUtils.defaultString(searchText)));
            criteria.add(disjunction);
        }

        if (itemsPerPage >= 0)
        {
            criteria.setMaxResults(itemsPerPage);
        }
        if (page > 1)
        {
            criteria.setFirstResult((page - 1) * itemsPerPage);
        }

        return criteria.execute();
    }

    /**
     * @param repo
     * @param startnode
     * @return
     */
    protected Criteria createCriteria(String repo, String startnode)
    {
        Criteria criteria = JCRCriteriaFactory
            .createCriteria()
            .setWorkspace(repo)
            .addOrder(Order.desc("@jcr:score"))
            .add(Restrictions.eq("@jcr:primaryType", "mgnl:content"));

        if (startnode != null)
        {
            criteria.setBasePath(startnode);
        }
        return criteria;
    }

    private Collection<AdvancedResultItem> collectCollectionFromResult(AdvancedResult result)
    {

        ResultIterator<AdvancedResultItem> items = result.getItems();
        ArrayList<AdvancedResultItem> list = new ArrayList<AdvancedResultItem>();

        CollectionUtils.addAll(list, items);
        return list;
    }

}
